/*
 mpg123/mpgaud/mpgaud.c
  snddec and Classic Sound Manager layer for MPEG audio output

  Revised: 2019.02.15
   Author: Adam J. P.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#if TARGET_OS_MAC
 #define SAMPLES(c) (c)->header.samplePtr
 #define NUMFRAMES(c) (c)->header.numFrames
 #define CHCOUNT(c) (c)->header.numChannels
 #define TASK(c) (c)->task.tmAddr
 #define iParam2 int32
#else
 #define SAMPLES(c) (c)->samplePtr
 #define NUMFRAMES(c) (c)->numFrames
 #define CHCOUNT(c) (c)->chCount
 #define TASK(c) (c)->task
 #define iParam2 void*
 #define sampledSynth 5
 #define initStereo sndInitStereo
 #define initMono 0
 #define flushCmd sndFlushCmd
 #define quietCmd sndQuietCmd
 #include "my_cseries.h"
 #include "../../../snddec/snddec.h"
 #define SndChannel sndChannel
 #define SndCommand sndCommand
 #define SndDisposeChannel(c,q) sndDisposeChannel(c)
 #define SndDoImmediate(ch,cmd) sndDoCommand(ch,cmd,TRUE)
short SndNewChannel(sndChannel **chan,
	short synth,
	long init,
	void *unused);
 #include <poppack.h>
#endif

#include <string.h>
#include "mpgaud.h"
#include "mevent.h"

/* in milliseconds */
#define AUTO_SCHEDULE_TIME	2 /* 1ms is too demanding */
/* this may lock up the system playing in break-up on inferior systems */
/* better for the program to just routinely call mpgaud_audiobuffer as usual */
//#define AUTO_RECOVER

/* ---- globals */

#if TARGET_OS_MAC
static TimerUPP		autoloaderUPP;
static SndCallBackUPP	playcallbackUPP;
static SndCallBackUPP	resetcallbackUPP;
#else
 #undef pascal
 #define pascal
extern timerThread	*global_timer_thread;
#endif

/* ---- private prototypes */

static pascal void playcallback(SndChannel *sndchannel,
	SndCommand *cmd);

/* ---- private code */

static int loader(mpgaud_channel *chan)
{
#if TARGET_OS_MAC
	FSVB		*vb;
#endif
	uint8		*bx;
	size_t		pcm_count;
	int		r;

	if (chan->bufindex>=chan->bufcount)
		return 0; /* full */
#if TARGET_OS_MAC
	vb = (FSVB*)chan->stream->handle->bitstream->file_handle;
	if (vb->busy)
		return vb->ioParam.ioResult<0 ? -1:4;
	vb->busy|= 2;
#else
	/* not entirely collision-proof, but mutexes cannot be used on a global timer thread */
	/* (maybe that should be changed); 2016.08.27 */
	if (chan->busy|(chan->channel->actFlags&chBusy))
		return 4;
	++chan->busy;
#endif
	bx = 0;
	if ((r=mpg123_decode_frame(chan->stream,NULL,&bx,&pcm_count))==MPG123_OK)
		memcpy(chan->buftotal+SAMPLES(chan),bx,pcm_count);
#if TARGET_OS_MAC
	vb->busy&=~2;
#else
	--chan->busy;
#endif
	if (r!=MPG123_OK)
		{chan->done = 1; return -ABS(r);}
	++chan->bufindex, chan->buftotal+= pcm_count;
	return 1;
}

/* total loading must be done well ahead of playback, of course */

#ifdef env68k
 #pragma parameter __A1 GetA1()
 extern void *GetA1(void) = {0x2009};
 #pragma parameter __D0 SetA1(__A1)
 extern long SetA1(void*) = {0x2009};
 #pragma parameter __D0 GetA5()
 extern long GetA5(void) = {0x200D};

static pascal void autoloader(...)
{
	mpgaud_channel	*chan = GetA1();
	long		saveA5;
#elif TARGET_OS_MAC
static pascal void autoloader(TMTask *param)
{
	mpgaud_channel	*chan = (mpgaud_channel*)param;
#else
static void autoloader(timerEntry *param)
{
	mpgaud_channel	*chan = param->userInfo;
#endif
	long		r;

	if (!chan->channel->userInfo)
		return;
#ifdef env68k
	saveA5 = SetA5(chan->myA5);
#endif
	r = loader(chan);
#ifdef env68k
	SetA5(saveA5);
#endif
	if (r<0)
		;//chan->channel->userInfo = 0; - ruins the end
#ifdef AUTO_RECOVER
	else if (r && NUMFRAMES(chan))
		if (chan->channel->userInfo)
		#if TARGET_OS_MAC
			PrimeTime((QElemPtr)&chan->task,AUTO_SCHEDULE_TIME);
		#else
			PrimeTimer(global_timer_thread,chan->task,AUTO_SCHEDULE_TIME*1000);
		#endif
		else
		{ /* we can save it! */
			chan->channel->userInfo = (long)chan;
			playcallback(chan->channel,NULL);
		}
#else
	else if (r && chan->channel->userInfo)
	#if TARGET_OS_MAC
		PrimeTime((QElemPtr)&chan->task,AUTO_SCHEDULE_TIME);
	#else
		PrimeTimer(global_timer_thread,chan->task,AUTO_SCHEDULE_TIME*1000);
	#endif
#endif
}

static pascal void resetcallback(SndChannel *sndchannel,
	SndCommand *cmd)
{
#pragma unused(cmd)
	sndchannel->userInfo = 0;
}

static pascal void playcallback(SndChannel *sndchannel,
	SndCommand *cmd)
{
#pragma unused(cmd)
	mpgaud_channel	*chan;

	if ((chan=(mpgaud_channel*)sndchannel->userInfo)==NULL)
		return;
	if (chan->buftotal)
	{
		SndCommand	sndcmd;

		NUMFRAMES(chan) = (chan->buftotal/chan->chSize)>>(CHCOUNT(chan)-1);
	#if TARGET_OS_MAC
		sndcmd.cmd = bufferCmd;
		sndcmd.param1 = 0;
		sndcmd.param2 = (long)&chan->header;
		SndDoImmediate(chan->channel,&sndcmd);

		sndcmd.cmd = callBackCmd;
		sndcmd.param2 = 0;
		SndDoCommand(chan->channel,&sndcmd,true);
	#else
		sndcmd.cmd = sndRawCmd;
		sndcmd.param1 = NUMFRAMES(chan);
		sndcmd.param2 = chan->samplePtr;
		sndDoCommand(chan->channel,&sndcmd,0);
		sndcmd.cmd = sndCallBackCmd;
		sndcmd.param2 = (void*)playcallback;
		sndDoCommand(chan->channel,&sndcmd,0);
	#endif
		/* swap buffers, and reset counters */
		if (SAMPLES(chan)==(void*)chan->buffer)
			SAMPLES(chan) = (void*)chan->midptr;
		else
			SAMPLES(chan) = (void*)chan->buffer;
		chan->bufindex = 0, chan->buftotal = 0;
	} else
		sndchannel->userInfo = 0; /* need samples! */
	if (TASK(chan))
	{
#ifdef env68k
		SetA1(&chan->task);
		autoloader();
#elif TARGET_OS_MAC
		autoloader(&chan->task);
#else
		autoloader(chan->task);
#endif
	}
}

/* ---- public code */

int mpgaud_audiobuffer(mpgaud_channel *chan)
{
	int		r;

	if (!chan || !SAMPLES(chan) || !chan->channel)
		return 0;
	if (chan->channel->userInfo) /* buffered, playing */
	{
		if (TASK(chan))
			return 1; /* loading at interrupt time */
		if (chan->done)
			__debugbreak();
		if ((r=loader(chan))<0)
			return -1; /* finished (or error) */
		return r;
	}
	//if (chan->bufindex>chan->bufcount)
	//	chan->bufindex = 0;
	if (!NUMFRAMES(chan) && chan->bufindex<chan->bufcount)
	{
		if ((r=loader(chan))<0)
			return -1; /* finished (or error) */
		return 0; /* one side must be full before play */
	}
	/* play if any completion - filled or final - has occurred */
	chan->channel->userInfo = (iParam2)chan;
	playcallback(chan->channel,NULL);
	/* can't expect low latency on all systems... */
	return loader(chan)<0 ? -1:2;
} /* mpgaud_audiobuffer */

int mpgaud_audiostop(mpgaud_channel *chan)
{
	SndCommand	cmd;

	if (!chan || !chan->channel)
		return -1;
	NUMFRAMES(chan) = chan->buftotal = chan->bufindex = 0;
	chan->channel->userInfo = 0;
	cmd.cmd = quietCmd;
	cmd.param1 = 0;
	cmd.param2 = 0;
	SndDoImmediate(chan->channel,&cmd);
	cmd.cmd = flushCmd;
	SndDoImmediate(chan->channel,&cmd);
	return 0;
} /* mpgaud_audiostop */

int mpgaud_audioclose(mpgaud_channel *chan)
{
	int		c = 0;

	if (!chan || !chan->bufcount) /* bufcount must be valid */
		return -1;
	/* stop any interrupt-level playback and buffering */
	mpgaud_audiostop(chan);
	if (TASK(chan))
	{
	#if TARGET_OS_MAC
		RmvTime((QElemPtr)&chan->task);
	#else
		RemoveTimer(global_timer_thread,chan->task);
	#endif
		TASK(chan) = NULL;
	}
	if (chan->buffer)
		{free(chan->buffer); chan->buffer = NULL, ++c;}
	if (chan->channel)
	{ /* passing false for quiet may hang... */
		if (!SndDisposeChannel(chan->channel,true))
			++c;
	#if TARGET_OS_MAC
		DisposePtr((Ptr)chan->channel);
	#endif
		chan->channel = NULL; ++c;
	}
	chan->bufcount = 0;
	free(chan);
	return c;
} /* mpgaud_audioclose */

int mpgaud_audioinit(mpgaud_channel **mchan,
	mpg123_handle *mpgh,
	int16 bufcount) /* negative for async playback */
{
 #if TARGET_OS_MAC
	SndCallBackUPP	cb;
 #else
	sndFormat	fmt;
 #endif
	mpgaud_channel	*chan;
	int		err, chanc, enc;

	if (!mchan || !mpgh || !bufcount)
		return -1;
	if ((chan=*mchan)==NULL)
		if ((*mchan=chan=(mpgaud_channel*)calloc(sizeof(mpgaud_channel),1))==NULL)
			return -2;
	mpg123_getformat(mpgh,&chan->rate,&chanc,&enc);
	if (chanc<=0)
		{err =-1; goto cerr;}
	if (chanc==3) /* 'channels' may be represented as mono/stereo flag bits */
		chanc = 2;
	CHCOUNT(chan) = chanc;
	/* set up the header */
	chan->chSize = mpg123_encsize(enc);
	if (!chan->buffer)
	{
		int32		maxbuf;

		chan->bufcount = bufcount<0 ? -bufcount:bufcount;
		maxbuf = chan->chSize*chan->bufcount*MPG_MAX_SAMPLES*CHCOUNT(chan);
		if ((chan->buffer=calloc(2,maxbuf))==NULL)
			{err =-2; goto cerr;}
		chan->midptr = chan->buffer+maxbuf;
	}
	SAMPLES(chan) = (void*)chan->buffer;
 #if TARGET_OS_MAC
	chan->header.sampleRate = chan->rate<<16;
	chan->header.sampleSize = chan->chSize<<3;
	chan->header.encode = cmpSH;
	chan->header.baseFrequency = kMiddleC;
	/* assumes 16-bit until Classic compatible rewrite */
	chan->header.format = BIG_ENDIAN ? k16BitBigEndianFormat:k16BitLittleEndianFormat;
	chan->header.samplePtr = (char*)chan->buffer;
 #else
	memset(&fmt,0,sizeof(fmt));
	fmt.chanCount = CHCOUNT(chan);
	fmt.codec = enc&MPG123_ENC_FLOAT_64 ? (BIG_ENDIAN ? CODEC_PCM_BE_F64:CODEC_PCM_LE_F64)
		:enc&MPG123_ENC_FLOAT_32 ? (BIG_ENDIAN ? CODEC_PCM_BE_F32:CODEC_PCM_LE_F32)
		:chan->chSize>2 ? (BIG_ENDIAN ? CODEC_PCM_BE_S32:CODEC_PCM_LE_S32)
		:chan->chSize>1 ? enc&MPG123_ENC_SIGNED
		? (BIG_ENDIAN ? CODEC_PCM_BE_S16:CODEC_PCM_LE_S16)
		:(BIG_ENDIAN ? CODEC_PCM_BE_S16:CODEC_PCM_LE_S16)
		:enc==MPG123_ENC_ULAW_8 ? CODEC_ULAW
		:enc==MPG123_ENC_ALAW_8 ? CODEC_ALAW
		:enc&MPG123_ENC_SIGNED ? CODEC_PCM_DEFAULT:CODEC_PCM_U8;
	fmt.rate = chan->rate;
	fmt.depth = chan->chSize<<3;
	/* no immediate codec values for u24/u32, so they're handled here */
	if (fmt.depth==32 && !(enc&(MPG123_ENC_FLOAT|MPG123_ENC_SIGNED)))
		fmt.type|= STYPE_ISGN;
 #endif
	/* set up the counters */
	chan->bufindex = 0;
	chan->buftotal = NUMFRAMES(chan) = 0;
 #ifdef env68k
	chan->myA5 = GetA5();
 #endif
	if (bufcount<0)
	{
	#if TARGET_OS_MAC
		if (!playcallbackUPP)
			playcallbackUPP = NewSndCallBackUPP(playcallback);
		cb = playcallbackUPP;
		if (!autoloaderUPP)
			autoloaderUPP = NewTimerUPP(autoloader);
		chan->task.tmAddr = autoloaderUPP;
	#endif
	} else
	{
	#if TARGET_OS_MAC
		if (!resetcallbackUPP)
			resetcallbackUPP = NewSndCallBackUPP(resetcallback);
		cb = resetcallbackUPP;
		if (chan->task.tmAddr)
			{RmvTime((QElemPtr)&chan->task); chan->task.tmAddr = NULL;}
	#else
	 #define cb resetcallback
		if (chan->task)
			{RemoveTimer(global_timer_thread,chan->task); chan->task = NULL;}
	#endif
	}
	chan->stream = mpgh;
	/* create the sound channel */
	if (!chan->channel)
	{
	#if TARGET_OS_MAC
		if ((chan->channel=(SndChannelPtr)NewPtrClear(sizeof(SndChannel)))==NULL)
			return mpgaud_ERR_MEM;
		chan->channel->qLength = 2;
	#endif
		if ((err=SndNewChannel(&chan->channel,sampledSynth,CHCOUNT(chan)==2 ?
				initStereo:initMono,(void*)cb))!=0)
		{
			err = 0;
		cerr:	mpgaud_audioclose(chan);
			*mchan = NULL;
			return err;
	#if TARGET_OS_MAC
		} else if (chan->task.tmAddr)
			InsTime((QElemPtr)&chan->task);
	#else
		} else
		{
			sndCommand	cmd;

			cmd.cmd = sndFormatCmd;
			cmd.param1 = 0;
			cmd.param2 = &fmt;
			sndDoCommand(chan->channel,&cmd,1);
			if ((chan->task=InstallTimer(global_timer_thread,autoloader))!=NULL)
				chan->task->userInfo = chan;
		}
	#endif
	}
	return 1;
} /* mpgaud_audioinit */
